import os, zipfile
from contextlib import contextmanager
from pytz import timezone
from datetime import datetime, timedelta
from threading import Lock

from .height import cleanDateTimesWithHeights
from .rain import appendedRainReadings
from ..config import MIN_ALTITUDE_DEG, IMAGE_DIR, CEILOMETER_DIR, DATASET_DIR, RAIN_ZIP_FILE
from ..image_preproc import decode_imagebuffer

def gatherDatapoints(image_dir=IMAGE_DIR, ceilometer_dir=CEILOMETER_DIR, fix_infinity=False, rain_data=True):
    # filtered data_points: list of (datetime, height)
    height_points = cleanDateTimesWithHeights(image_dir, ceilometer_dir, fix_infinity=fix_infinity)

    if rain_data:
        # data_points: list of (datetime, height, rain_dict)
        rain_zip_file = os.path.join(DATASET_DIR, RAIN_ZIP_FILE)
        data_points = appendedRainReadings(height_points, rain_zip_file)

    return data_points



def datetime_to_image(date_time, image_zip_dir, ozips):
    zipFile = os.path.join(image_zip_dir, dateTimeToZipFileName(date_time))
    imgFileName = dateTimeToImageName(date_time)

    # read image from zip (threadsave)
    with ozips.lock:
        img_buffer = ozips.imgBufferFromZip(zipFile, imgFileName)

    # decode raw image
    return decode_imagebuffer(img_buffer)


@contextmanager
def openZips(maxCount):
    """allow maxCount ZipFiles to stay open to speed up random access"""
    class OpenedZips():
        def __init__(self, maxZipsOpen):
            self.maxOpenZipFiles = 10
            self.openedZipFiles = []
            self.lock = Lock()

        def imgBufferFromZip(self, zipFileName, imgName):
            """opens zip with zipFileName if not already open and reads imgName file"""
            #openZf is tuple (name of zip file, opened file)
            openZf = next((f for f in self.openedZipFiles if f[0] == zipFileName), None)
            if openZf:
                self.openedZipFiles.remove(openZf)
            else:
                if len(self.openedZipFiles) >= self.maxOpenZipFiles:
                    self.openedZipFiles.pop()[1].close()
                openZf = (zipFileName, zipfile.ZipFile(zipFileName, 'r'))
            
            self.openedZipFiles.append(openZf)
            return openZf[1].read(imgName)

        def close(self):
            """close all opened zip files"""
            for openZf in self.openedZipFiles:
                openZf[1].close()
            self.openedZipFiles = []

    ozips = OpenedZips(maxCount)
    yield ozips
    ozips.close()


### time zone util



def dateTimeToImageName(dt_utc, utcp=1):
    dtShifted = dt_utc + timedelta(hours=utcp)
    return dtShifted.strftime("Image_%Y%m%d_%H%M%S_UTCp" + str(utcp) + ".jpg")

def dateTimeToZipFileName(dt_utc, camera_name='Wkm', utcp=1):
    dtShifted = dt_utc + timedelta(hours=utcp)
    return dtShifted.strftime("Images_" + camera_name + "_%Y%m%d.zip")
