import os, glob, re, zipfile, csv
from dateutil.parser import parse as dateparse
from pytz import timezone
from datetime import datetime, timedelta


from itertools import groupby
from concurrent.futures import ThreadPoolExecutor, wait


from ..config import MIN_ALTITUDE_DEG, IMAGE_DIR, CEILOMETER_DIR, DATASET_DIR

from .sunlocation import getSunAltitude

## HEIGHT READING

def cleanDateTimesWithHeights(image_dir, ceilometer_dir, fix_infinity=True):
    """filter out datetimes with low altitude of the sun"""
    # do not look up csv values for night images
    day_filter = lambda x:getSunAltitude(x) >= MIN_ALTITUDE_DEG

    data_points = allDateTimesWithHeights(image_dir, ceilometer_dir, day_filter)
    # replace infinity height readings with number
    if data_points:
        if fix_infinity:
            return fixInfinityReading(data_points)
        return data_points
    return []


def allDateTimesWithHeights(image_dir, ceilometer_dir, datetime_filter_func=lambda x:True):
    """assumtion: each zip file contains images with name Image_YYYYMMDD_HHMMSS_UTCp1.jpg"""
    if not (os.path.isdir(image_dir) and os.path.isdir(ceilometer_dir)):
        raise ValueError('Gathering directory does not exist! "{}" or "{}"'.format(image_dir, ceilometer_dir))

    imageZips = list(filter(zipfile.is_zipfile, glob.glob(os.path.join(image_dir, '*.zip'))))

    data_points = []
    with ThreadPoolExecutor(max_workers=4) as executor:
        #speed up by csv lookup parallization
        dp_lookup = lambda iz: list(allDateTimesWithHeightsInZip(iz, ceilometer_dir, datetime_filter_func))
        for iz, dps in zip(imageZips, executor.map(dp_lookup, imageZips)):
            #inform about findings per zip file
            print('Found %d height readings for: %s' % (len(dps), iz))
            data_points.extend(dps)
    return data_points


def all_ceilometer_readings(ceilometer_dir):
    filenames = all_ceilometer_filenames(ceilometer_dir)
    with ThreadPoolExecutor(max_workers=20) as executor:
        filename_to_heights = lambda n: list(all_csv_ceilometer_readings(n))

        for filename, reading_tuples in zip(filenames, executor.map(filename_to_heights, filenames)):
            print('ceilometer readings from {}'.format(filename))

            for rowTime, reading in reading_tuples:
                yield rowTime, reading


def all_csv_ceilometer_readings(csv_filename):
    with open(csv_filename) as csvfile:
        reader = csv.DictReader(csvfile, delimiter=';')
        for row in reader:
            #row time in csv is in ascending order
            rowTime = stringDateTimeParse(row['DATE'], row['TIME'], row['TIMEZONE'])
            #get lowest reading from u/m/o
            reading = getLowestRowCloudReading(row)
            if reading:
                yield (rowTime, reading)

def findBestCeilometerReadings(searchTimes, csv_filename):
    fileTimeSpan = timedelta(days=1)
    maxDeviation = timedelta(seconds=90)

    #initilize closest result with maxmial deviation
    closestResults = {st:([], maxDeviation) for st in searchTimes}
    for rowTime, reading in all_csv_ceilometer_readings(csv_filename):

        # search for closer readings than in closest results
        for searchTime in list(closestResults.keys()):
            _, minDeviation = closestResults[searchTime]
            timeDiff = rowTime - searchTime
            # allowed time differences are in interval [-1, 1] 
            searchIntervalLoc = timeDiff / maxDeviation

            if searchIntervalLoc > 1:
                yield (searchTime, closestResults[searchTime][0])
                #interval is over: no closer results will follow
                # search is done
                del closestResults[searchTime]
            elif abs(searchIntervalLoc) <= 1:
                # timeDiff should be closer than previous closest 
                # if the closest result is 0 (no better result can be found)
                if timedelta(0) != minDeviation and abs(timeDiff / minDeviation) < 1:
                    
                    closestResults[searchTime] = (reading, timeDiff)
                    
    for st, cres in closestResults.items():
        yield st, cres[0]

def getLowestRowCloudReading(row):
    """Returns ceilometer cloud height reading (used detection status: 0-5)"""
    status = row['CTDS']
    if status:
        d = int(status)
        if d == 0 or d ==  5:
            # no cloud detected
            return float("inf")
        elif d == 4:
            labels = ['CTVI','CTHS']
        else:
            labels = ['CTWBU'] # 'CTWBM', 'CTWBO' (always higher)
        return min([int(row[label]) for label in labels])

def searchCeilometerReadings(allSearchTimes, ceilometer_dir):
    #get csv file names by date time (serachtimes)
    time_to_filename = lambda x:getCeilometerFileNameByDate(ceilometer_dir, x)

    for csv_filename, searchTimes in groupby(allSearchTimes, time_to_filename):
        #search date times in correct csv file
        for searchTime, reading in findBestCeilometerReadings(searchTimes, csv_filename):
            if reading:
                yield (searchTime, reading)
            else:
                print('no reaing found for', searchTime, 'in', csv_filename)

def allDateTimesWithHeightsInZip(images_zip, ceilometer_dir, datetime_filter_func):
    """Returns all (date_time, height reading) tuples based on date_times in zip file

    Args:
      images_zip: zip file of cloud images
      ceilometer_dir: directory with csv subdirectories
      datetime_filter_func: filter function for datetimes
    Returns:
      All found tuples (date_time, height reading)
    """
    imageNames = []
    with zipfile.ZipFile(images_zip, 'r') as zfile:
        #all search times are based on the file names in the zip file
        imageNames = zfile.namelist()
        searchTimes = [imageFileNameToDateTimeUTC(iname) for iname in imageNames]
    return searchCeilometerReadings(filter(datetime_filter_func, searchTimes), ceilometer_dir)
            

def all_ceilometer_filenames(ceilometer_dir):
    return glob.glob(os.path.join(ceilometer_dir, '*', '*.ct'))

def getCeilometerFileNameByDate(ceilometer_dir, dt):
    """Returns the path of the csv file based on a datetime"""
    yearMonth = dt.strftime('%Y%m')
    monthDay = dt.strftime('%m%d')
    #example: ct/201712/1231.ct
    return os.path.join(ceilometer_dir, yearMonth, monthDay + '.ct')


def fixInfinityReading(data_points):
    """Replace inifity height with (max height * 2)"""
    # find max non inf height
    maxCloudHeight = max(filter(lambda x: x < float("inf"), map(lambda x: x[1], data_points)))
    print('maximum cloud height:', maxCloudHeight)

    # no cloud reading is interpreted as: maxCloudHeight * 2
    return list(map(lambda x: (x[0], int(min(maxCloudHeight*2, x[1]))), data_points))


## datetime util

def stringDateTimeParse(date_str, date_time_str, timezone_str):
    return dateparse('%s %s %s' % (date_str, date_time_str, timezone_str))


def imageFileNameToDateTimeUTC(imageFileName):
    result = re.match('Image_(.*)_UTCp(\\d+).jpg', imageFileName)
    if result:
        d = datetime.strptime(result.group(1),'%Y%m%d_%H%M%S') - timedelta(hours=int(result.group(2)))
        return timezone('UTC').localize(d)
    return None