import re, zipfile
from datetime import datetime, timedelta
from pytz import timezone
import numpy as np

from ..config import DATASET_DIR, RAIN_ZIP_FILE

## RAIN READING
most_frequent = lambda a: np.argmax(np.bincount(a))

READ_OPT = {'RD':{'aggregate':most_frequent, 'around':(-30,0), 'dtype':np.int},
            'RR':{'aggregate':np.max, 'around':(-10,10), 'dtype':np.float},
            'GSD':{'aggregate':most_frequent,'around':(-5,5), 'dtype':np.int},
            'G':{'aggregate':np.mean, 'around':(-5,5), 'dtype':np.float}}

def appendedRainReadings(points, rain_zip):
    date_times = map(lambda x: x[0], points)
    rain_dicts = allAggregateRainReadings(date_times, rain_zip)
    return [p + (r,) for p, r in zip(points, rain_dicts)]

def allRainReadingDetectors(rain_zip):
    """Yields for each (RD, RR, GSD, G): detector as numpy array with 1 minute interval readings"""
    with zipfile.ZipFile(rain_zip, 'r') as zfile:
        for name, token, start_dt, end_dt in parseRainFileNames(zfile.namelist()):
            if token in READ_OPT:
                opt = READ_OPT[token]
                detector = np.fromstring(zfile.read(name), dtype=opt['dtype'], sep='\n')
                yield (token, detector, start_dt)

def allAggregateRainReadings(date_times, rain_zip):
    gather_funcs = {}
    detectors = {token:(detector, start_dt) for token, detector, start_dt in allRainReadingDetectors(rain_zip)}
        
    for dt in date_times:
        rd = {}
        for token, (detector, start_dt) in detectors.items():
            opt = READ_OPT[token]
            rd[token] = aggregateReadingAround(dt, start_dt, detector, 
                opt['aggregate'], opt['around'])
        # rain readings for each date_time e.g. {'RD':1, ..., 'G':300}
        yield rd


def parseRainFileNames(names):
    regex = re.compile('([A-z]+)_(\\d+)-(\\d+).txt')
    tz = timezone('Europe/Amsterdam')
    for name in names:
        result = regex.match(name)
        if result:
            token = result.group(1)
            start_dt = datetime.strptime(result.group(2),'%Y%m%d%H%M')
            end_dt = datetime.strptime(result.group(3),'%Y%m%d%H%M')
            yield name, token, tz.localize(start_dt), tz.localize(end_dt)


def aggregateReadingAround(dt, start_dt, detector, aggregate=np.max, around=(-1, 1)):
    """Aggregates a valid interval around the datetime from minute seperated detector values"""
    minute = int((dt - start_dt).total_seconds() // 60)
    a, b = around
    minute_start = max(minute+a, 0)
    minute_end = min(minute+b, len(detector) - 1)
    if minute_end > minute_start:

        region = detector[minute_start:minute_end]
        valid_region = region[region != 99999]
        if len(valid_region) > 0:
            return aggregate(valid_region)
    return 0