from .reading import gatherDatapoints, datetime_to_image, openZips, allRainReadingDetectors, all_ceilometer_readings
from .image_preproc import processed_image, encode_image
from .validate import load_datapoints, load_heightpoints, height_histo
import numpy as np
import cv2
import matplotlib.pyplot as plt
import os
from .config import IMAGE_DIR, CEILOMETER_DIR, DATASET_DIR, RAIN_ZIP_FILE


def do_histogram(value_name, values, mask, bins):
    vs = values[mask]
    plt.hist(vs, bins=bins)
    plt.title('{} for {} readings. {:.1%} of total'.format(value_name, 
        len(vs), 
        len(vs)/len(values)))


def height_histogram(heights):
    do_histogram('cloud heights < inf', heights, np.isfinite(heights), 200)
    plt.show()

def rain_sun_histogram(rain_quantities, sun_insolations):
    plt.subplot(211)
    do_histogram('rain_quantities > 0', rain_quantities, rain_quantities > 0, 10)

    plt.subplot(212)
    do_histogram('sun_insolations > 0', sun_insolations, sun_insolations > 0, 100)
    plt.show()


def show_random_height_histogram(ceilometer_dir):
    hps = load_heightpoints(ceilometer_dir)
    hist, bins = height_histo(hps)
    heights = np.array([h for t,h in hps])
    heights[np.isinf(heights)] = 8000
    height_histogram(heights)

    heights = random_heights(hist, bins, 50000)
    heights[np.isinf(heights)] = 8000
    height_histogram(heights)


def rain_anomalies(dps):
    for date, height, rain_dict in dps:
        rain_detect = rain_dict['RD']
        sun_detect = rain_dict['GSD']
        sun_insolation = rain_dict['G']
        if rain_detect == 1 and height > 5000:
            yield (date, height, rain_detect, sun_detect, sun_insolation)



def visualize_anomalies(image_dir, ceilometer_dir):
    dps = load_datapoints(image_dir, ceilometer_dir)
    print('loaded dps:', len(dps))

    with openZips(5) as ozips:
        for date, height, rain_detect, sun_detect, sun_insolation in rain_anomalies(dps):
            pp_img = datetime_to_image(date, image_dir, ozips)
            img = processed_image(pp_img)

            plt.subplot(121)
            plt.title('{}, Height:{} [R:{} S:{},{}]'.format(date, height, rain_detect, sun_detect, sun_insolation))
            plt.axis('off')
            plt.imshow(cv2.cvtColor(img, cv2.COLOR_BGR2RGB))

            plt.subplot(122)
            plt.axis('off')
            plt.imshow(cv2.cvtColor(pp_img, cv2.COLOR_BGR2RGB))
            plt.show()
            break



def visualize_filtered_histograms(image_dir, ceilometer_dir):
    dps = load_datapoints(image_dir, ceilometer_dir)
    rain_quantities = np.array([dp[2]['RR'] for dp in dps])
    sun_insolations = np.array([dp[2]['G'] for dp in dps])
    rain_sun_histogram(rain_quantities, sun_insolations)

    heights = np.array([dp[1] for dp in dps])
    height_histogram(heights)

def visualize_rain_sensitivity(rain_zip_file):
    for token, detector, start_dt in allRainReadingDetectors(rain_zip_file):
        if token == 'RR':
            rain_quantities = detector
        elif token == 'RD':
            rain_detections = detector

    near_zero = rain_quantities[rain_detections == 0]
    near_zero = near_zero[near_zero != 99999]
    bigger_zero = rain_quantities[rain_detections == 1]
    bigger_zero = bigger_zero[bigger_zero != 99999]

    plt.subplot(321)
    plt.title('rain_quantities[rain_detections == 0] for {} readings. mean:{}'.format(len(near_zero), np.mean(near_zero)))
    plt.hist(near_zero, bins=10)
    plt.subplot(322)
    nz2 = near_zero[near_zero > 0]
    plt.title('rain_quantities[rain_detections == 0] > 0 for {} readings. mean:{}'.format(len(nz2), np.mean(nz2)))
    plt.hist(nz2, bins=10)

    plt.subplot(323)
    plt.title('rain_quantities[rain_detections == 1] for {} readings. mean:{}'.format(len(bigger_zero), np.mean(bigger_zero)))
    plt.hist(bigger_zero, bins=10)
    plt.subplot(324)
    bz2 = bigger_zero[bigger_zero > 0]
    plt.title('rain_quantities[rain_detections == 1] > 0 for {} readings. mean:{}'.format(len(bz2), np.mean(bz2)))
    plt.hist(bz2, bins=10)

    quant = rain_detections[rain_quantities > 0]
    quant = quant[quant != 99999]
    noquant = rain_detections[rain_quantities == 0]
    noquant = noquant[noquant != 99999]

    plt.subplot(325)

    plt.title('rain_detections[rain_quantities > 0] for {} readings. mean:{}'.format(len(quant), np.mean(quant)))
    plt.hist(quant, bins=10)

    plt.subplot(326)
    plt.title('rain_detections[rain_quantities == 0] for {} readings. mean:{}'.format(len(noquant), np.mean(noquant)))
    plt.hist(noquant, bins=10)

    plt.show()

def visualize_histograms(ceilometer_dir, rain_zip_file):
    for token, detector, start_dt in allRainReadingDetectors(rain_zip_file):
        reg_mask = detector != 99999
        reg = detector[reg_mask]
        if token == 'RR':
            rain_quantities = reg
        elif token == 'G':
            sun_insolations = reg

    rain_sun_histogram(rain_quantities, sun_insolations)

    heights = np.array([height for rowTime, height in load_heightpoints(ceilometer_dir)])
    print('loaded heights:', heights.size)
    height_histogram(heights)


if __name__ == '__main__':
    image_dir = '..\\cloud_images'
    ceilometer_dir = '..\\ct'
    rain_zip_file = os.path.join(DATASET_DIR, RAIN_ZIP_FILE)
    #visualize_anomalies(image_dir, ceilometer_dir)
    #visualize_histograms(ceilometer_dir, rain_zip_file)
    visualize_filtered_histograms(image_dir, ceilometer_dir)
    #visualize_rain_sensitivity(rain_zip_file)
