import cv2
import argparse
import tensorflow as tf
import numpy as np
import zipfile as zf
import glob
import sys, os
import csv
import pickle

from datetime import datetime, timedelta, tzinfo
from itertools import groupby

from contextlib import contextmanager

import re
import random

import tensorflow as tf

def imageFileNameToDateTimeUTC(imageFileName):
	result = re.match('Image_(.*)_UTCp(\\d+).jpg', imageFileName)
	if result:
		return datetime.strptime(result.group(1),'%Y%m%d_%H%M%S') - timedelta(hours=int(result.group(2)))
	return None

def dateTimeToImageName(dt, utcp=1):
	dtShifted = dt + timedelta(hours=utcp)
	return dtShifted.strftime("Image_%Y%m%d_%H%M%S_UTCp" + str(utcp) + ".jpg")


def dateTimeToZipFileName(dt, camera_name='Wkm', utcp=1):
	dtShifted = dt + timedelta(hours=utcp)
	return dtShifted.strftime("Images_" + camera_name + "_%Y%m%d.zip")


def findBestCeilometerReadings(searchTimes, csvfileName):
	fileTimeSpan = timedelta(days=1)
	maxDeviation = timedelta(seconds=90)

	closestResults = {}
	with open(csvfileName) as csvfile:
		reader = csv.DictReader(csvfile, delimiter=';')
		for i, row in enumerate(reader):
			rowTime = datetime.strptime(row['DATE']+row['TIME'], '%d.%m.%Y%H:%M:%S')
			if i == 0:
				# filter searchTimes that can be found in this file
				searchTimes = list(filter(lambda st:abs((rowTime - st) / fileTimeSpan) < 1, searchTimes))
				closestResults = {st:([], maxDeviation) for st in searchTimes}
				if not closestResults: break

			for searchTime in list(closestResults.keys()):
				timeDiff = rowTime - searchTime
				# -1 to 1 
				searchIntervalLoc = timeDiff / maxDeviation
				if searchIntervalLoc > -1:
					# if the closest result is 0 (no better result can be found)
					if searchIntervalLoc < 1 and timedelta(0) != closestResults[searchTime][1]:
						# timeDiff should be closer than previous closest 
						if abs(timeDiff / closestResults[searchTime][1]) < 1:
							#get lowest reading from u/m/o
							reading = getLowestRowCloudReading(row)
							if reading:
								closestResults[searchTime] = (reading, timeDiff)
								
					else:
						#interval is over: no close results will follow
						yield (searchTime, closestResults[searchTime][0])
						del closestResults[searchTime]
	for st, cres in closestResults.items():
		yield st, cres[0]

def getLowestRowCloudReading(row):
	#used detection status: 0,1,2,3
	status = row['CTDS']
	if status:
		d = int(status)
		if d == 0 or d ==  5:
			return float("inf")
		elif d == 4:
			labels = ['CTVI','CTHS']
		else:
			labels = ['CTWBU'][:d] # 'CTWBM', 'CTWBO' (always higher)
		return min([int(row[label]) for label in labels])

def searchCeilometerReadings(allSearchTimes):
	for csv_filename, searchTimes in groupby(allSearchTimes, getCeilometerFileNameByDate):
		for searchTime, reading in findBestCeilometerReadings(searchTimes, csv_filename):
			print(searchTime, reading)
			yield (searchTime, reading)

def allLabeledImagesInZip(iz):
	
	print('searching ceilo labels for:', iz)
	imageNames = []
	with zf.ZipFile(iz, 'r') as zfile:
		imageNames = zfile.namelist()
		searchTimes = [imageFileNameToDateTimeUTC(iname) for iname in imageNames]
		for imageDT, reading in searchCeilometerReadings(searchTimes):
			yield (imageDT,reading)

def _int64_feature(value):
	return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))


def _bytes_feature(value):
	return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))

def processImageData(img_buffer):
	# 1920x1920 to 480x480
	image = cv2.resize(cv2.imdecode(np.frombuffer(img_buffer, np.uint8), 1), (480,480))
	width = image.shape[0]
	height = image.shape[1]
	depth = image.shape[2]

	# more preprocessing...
	_, encodeArray = cv2.imencode('.png', image, [cv2.IMWRITE_PNG_COMPRESSION, 6])
	return width, height, depth, encodeArray


def getCeilometerFileNameByDate(dt):
	ceilometerDir = '../ct/'
	yearMonth = dt.strftime('%Y%m')
	monthDay = dt.strftime('%m%d')
	return os.path.join(ceilometerDir, yearMonth, monthDay + '.ct')





@contextmanager
def openZips(maxCount):
	# allow maxCount ZipFiles to stay open to speed up random access
	class OpenedZips():
		def __init__(self, maxZipsOpen):
			self.maxOpenZipFiles = 10
			self.openedZipFiles = []

		def imgBufferFromZip(self, zipFileName, imgName):
			openZf = next((f for f in self.openedZipFiles if f[0] == zipFileName), None)
			if openZf:
				self.openedZipFiles.remove(openZf)
			else:
				if len(self.openedZipFiles) >= self.maxOpenZipFiles:
					self.openedZipFiles.pop()[1].close()
				openZf = (zipFileName, zf.ZipFile(zipFileName, 'r'))
			
			self.openedZipFiles.append(openZf)
			return openZf[1].read(imgName)

		def close(self):
			for openZf in self.openedZipFiles:
				openZf[1].close()
			self.openedZipFiles = []

	ozips = OpenedZips(maxCount)
	yield ozips
	ozips.close()


def writeToTFRecordsFile(writer, data_points):
	with openZips(10) as ozips:
		pointCount = len(data_points)

		for i, (date_time, height_reading) in enumerate(data_points):
			imgFileName = dateTimeToImageName(date_time)
			zipFile = os.path.join(IMAGE_DIR, dateTimeToZipFileName(date_time))
			img_buffer = ozips.imgBufferFromZip( zipFile, imgFileName)

			width, height, depth, array = processImageData(img_buffer)
			example = tf.train.Example(features=tf.train.Features(feature={
				'timestamp': _int64_feature(int(date_time.timestamp())),
				'width': _int64_feature(width),
				'height': _int64_feature(height),
				'depth': _int64_feature(depth),
				'label': _int64_feature(height_reading),
				'image_encoded': _bytes_feature(array.tostring())}))
			writer.write(example.SerializeToString())

			if i % int(pointCount / 200) == 0:
				print('Written %d examples: (%.2f %%)' % (i, 100 * i /pointCount))





IMAGE_DIR = '../cloud_images/'

def uniformDateTimesWithHeights():
	# assumtion: each zip file contains images with name Image_YYYYMMDD_HHMMSS_UTCp1.jpg
	imageZips = glob.glob(os.path.join(IMAGE_DIR, '*.zip'))

	data_points = []
	for iz in imageZips:
		data_points.extend(allLabeledImagesInZip(iz))

	# find max non inf height  
	maxCloudHeight = max(filter(lambda x: x < float("inf"), map(lambda x: x[1], data_points)))
	print('maximum cloud height:', maxCloudHeight)

	# no cloud reading is interpreted as: maxCloudHeight * 2
	data_points = list(map(lambda x: (x[0], int(min(maxCloudHeight*2, x[1]))), data_points))

	print('shuffling data points...')
	random.shuffle(data_points)

	return data_points


def main(argv):
	
	dt_file = os.path.join(FLAGS.directory, 'dateWithHeight.dtpts')
	if os.path.isfile(dt_file):
		with open(dt_file, 'rb') as f:
			data_points = pickle.load(f)
	else:
		data_points = uniformDateTimesWithHeights()

	with open(dt_file, 'wb') as f:
		pickle.dump(data_points, f)

	### TODO filter out low sun positions


	pointCount = len(data_points)

	print('Total count of data points:', pointCount)


	# write data_points to train and validation set
	trainFilename = os.path.join(FLAGS.directory, 'train.tfrecords')
	print('Writing training set:', trainFilename)
	writerTrain = tf.python_io.TFRecordWriter(trainFilename)

	writeToTFRecordsFile(writerTrain, data_points[pointCount//2:])
	writerTrain.close()
	

	valFilename = os.path.join(FLAGS.directory, 'validation.tfrecords')
	print('Writing validation set:', trainFilename)
	writerValidation = tf.python_io.TFRecordWriter(valFilename)

	writeToTFRecordsFile(writerValidation, data_points[:pointCount//2])
	writerValidation.close()





if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument(
		  '--directory',
		  type=str,
		  default='output',
		  help='Directory to download data files and write the converted result'
	)
	parser.add_argument(
	'--validation_size',
	type=int,
	default=100,
	help="""\
	Number of examples to separate from the training data for the validation
	set.\
	"""
	)
	FLAGS, unparsed = parser.parse_known_args()
	tf.app.run(main=main, argv=[sys.argv[0]] + unparsed)
	print('Done.')

